// myList.h
// Eine doppelt verkette Liste.
// Verkettung mit shared_ptr und weak_ptr.
// --------------------------------------------------
#ifndef MYLIST_H
#define MYLIST_H

#include <iostream>
#include <cassert>
#include <memory>
#include <iterator>
using namespace std;

template<class T>
class MyList {
    struct Element {               // Typ der Elemente
      T obj;
      shared_ptr<Element> next;
      weak_ptr<Element> prev;
      Element(const T& a = T()) : obj(a) {}
    };
    shared_ptr<Element> first;
    weak_ptr<Element> last;
    size_t size_val;
  public:
    MyList() : size_val(0) {}
                                  // Kopierkonstruktor:
    MyList( const MyList& list2) : size_val(0) { append( list2); }

                                  // Zuweisung:
    MyList& operator=( const MyList& list)
    {
        if( this != &list)        // Keine Selbstzuweisung.
        {  clear();  append( list);  }
        return *this;
    }

    size_t size() const  { return size_val; }
    bool empty() const { return size_val == 0; }

    void clear()
    {
        first.reset();        // Lscht alle Elemente!
        last.reset();
        size_val = 0;
    }

    void push_front(const T& obj)
    {
      shared_ptr<Element> ePtr(new Element(obj));
      if(first) {
        ePtr->next = first;    // Nicht leer => verketten
        first->prev = ePtr;
        first = ePtr;
      }
      else
         last = first = ePtr;
      ++size_val;
    }
    void pop_front()
    {
      if(size_val > 1)
      {                        // Mindestens zwei Elemente.
        first = first->next;   // Auf 2. Element setzen.
                               // => 1. Element wird zerstrt.
        first->prev.reset();   // Kein Vorgnger.
        --size_val;
      }
      else
        clear();
    }

    T& front() { assert(first); return first->obj; }

    void push_back(const T& obj)
    {
      shared_ptr<Element> ePtr(new Element(obj));
      if(!empty()) {
        ePtr->prev = last;    // Nicht leer => anhngen
        last.lock()->next = ePtr;
        last = ePtr;
      }
      else
         last = first = ePtr;
      ++size_val;
    }

    void pop_back()
    {
      if(size_val > 1)
      {                             // Mindestens zwei Elemente.
        last = last.lock()->prev;   // Zeigt auf vorletztes Element.
        last.lock()->next.reset();  // Kein Nachfolger,
                                    // letztes Element wird zerstrt.
        --size_val;
      }
      else
        clear();
    }

    T& back() { assert(first); return last.lock()->obj; }

    void append( const MyList& list2)
    {
      if( this != &list2)    // Quelle != Ziel
      {
        shared_ptr<Element> p = list2.first;
        while( p )
        {
            push_back(p->obj);
            p = p->next;
        }
      }
    }
    template<class U>
    friend ostream& operator<<( ostream& os, const MyList<U>& list);

    // Ein bidirektionaler Iterator:
    class Iterator;
    Iterator begin() { return Iterator(first,this); }
    Iterator end() { return Iterator(this); }
};

// Ausgabe ohne Iteratoren. Fr T muss der Operator << definiert sein.
template<class T>
inline ostream& operator<<( ostream& os, const MyList<T>& l)
{
    auto ptr = l.first;
    while( ptr )
    {
       os << '\t' << ptr->obj;
       ptr = ptr->next;
    }
    return os;
}

// --------------------------------------------------
// Ein Iterator fr MyList<T>
template<class T>
class MyList<T>::Iterator : public std::iterator<std::bidirectional_iterator_tag, T>
{
    weak_ptr<Element> ptr;
    MyList<T> *listPtr;            // Zugrunde liegende Liste.
  public:
    Iterator(MyList * const lp=0 ) : listPtr(lp) {}
    Iterator(const weak_ptr<Element>& p, MyList * const lp)
    : ptr(p), listPtr(lp) {}

    T& operator*() const {
        shared_ptr<Element> ep = ptr.lock();
        assert(ep);
        return ep->obj;
    }
    T* operator->() const {
        shared_ptr<Element> ep = ptr.lock();
        assert(ep);
        return &ep->obj;
    }
    Iterator& operator++() {         // preincrement
        shared_ptr<Element> ep = ptr.lock();
        if(ep)
            ptr = ep->next;
        return *this;
    }
    Iterator operator++(int) {       // postincrement
        Iterator tmp = *this;
        ++*this;
        return tmp;
    }
    Iterator& operator--() {
        shared_ptr<Element> ep = ptr.lock();
        if(ep)
        { assert(ep!=listPtr->first);  ptr = ep->prev; }
        else
        { assert(!listPtr->empty());  ptr = listPtr->last;  }
        return *this;
    }
    Iterator operator--(int) {       // postdecrement
        Iterator tmp = *this;
        --*this;
        return tmp;
    }
    bool operator==( const Iterator& i2) const
    {  return ptr.lock() == i2.ptr.lock(); }
    bool operator!=( const Iterator& i2) const
    {  return !(*this == i2); }
};

#endif    // MYLIST_H

